@@ -40,6 +40,7 @@ typedef enum which_t {
   tapdevice,
   allowconnect,
   pingtimeout,
+  keyexpire,
 } which_t;
 
 typedef struct config_t {
@@ -64,7 +65,7 @@ extern config_t *config;
 extern int debug_lvl;
 extern int timeout;
 
-extern config_t *add_config_val(config_t **, int, const char *);
+extern config_t *add_config_val(config_t **, int, char *);
 extern int read_config_file(const char *);
 extern const config_t *get_config_val(which_t type);
 
