@@ -35,11 +35,13 @@
 #include <unistd.h>
 
 #include <cipher.h>
+#include <utils.h>
 #include <xalloc.h>
 
 #include "conf.h"
 #include "encr.h"
 #include "net.h"
+#include "netutl.h"
 #include "protocol.h"
 
 int tap_fd = -1;
@@ -49,95 +51,12 @@ int total_tap_out = 0;
 int total_socket_in = 0;
 int total_socket_out = 0;
 
+time_t last_ping_time = 0;
+
 /* The global list of existing connections */
 conn_list_t *conn_list = NULL;
 conn_list_t *myself = NULL;
 
-/*
-  look for a connection associated with the given vpn ip,
-  return its connection structure
-*/
-const conn_list_t *lookup_conn(ip_t ip)
-{
-  conn_list_t *p = conn_list;
-
-  for(p = conn_list; p != NULL; p = p->next)
-    if((ip & p->vpn_mask) == (p->vpn_ip & p->vpn_mask))
-      return p;
-
-  return NULL;
-}
-
-/*
-  free a conn_list_t element and all its pointers
-*/
-void free_conn_element(conn_list_t *p)
-{
-  if(p->hostname)
-    free(p->hostname);
-  if(p->pp)
-    free(p->pp);
-  free(p);
-}
-
-/*
-  remove all marked connections
-*/
-void prune_conn_list(void)
-{
-  conn_list_t *p, *q = NULL;
-
-  for(p = conn_list; p != NULL; )
-    {
-      if(p->status.remove)
-	{
-	  if(q)
-	    {
-	      q->next = p->next;
-	      free_conn_element(p);
-	      p = q->next;
-	    }
-	  else
-	    {
-	      conn_list = q = p->next;
-	      free_conn_element(p);
-	      p = q;
-	    }
-	}
-      else
-	p = p->next;
-    }
-}
-
-/*
-  creates new conn_list element, and initializes it
-*/
-conn_list_t *new_conn_list(void)
-{
-  conn_list_t *p = xmalloc(sizeof(conn_list_t));
-
-  memset(p, '\0', sizeof(conn_list_t));
-  return p;
-}
-
-void destroy_conn_list(void)
-{
-  conn_list_t *p, *q;
-
-  for(p = conn_list; p != NULL; )
-    {
-      if(p->hostname)
-	free(p->hostname);
-      if(p->pp)
-	free(p->pp);
-      q = p;
-      p = p->next;
-      free(q);
-    }
-
-  conn_list = NULL;
-}
-
 /*
   strip off the MAC adresses of an ethernet frame
 */
@@ -161,18 +80,171 @@ void add_mac_addresses(vpn_packet_t *p)
   p->len += 12;
   tmp[0] = tmp[6] = 0xfe;
   tmp[1] = tmp[7] = 0xfd;
-  *((ip_t*)(&tmp[2])) = *((ip_t*)(&tmp[26]));
-  *((ip_t*)(&tmp[8])) = *((ip_t*)(&tmp[30]));
+  *((ip_t*)(&tmp[2])) = *((ip_t*)(&tmp[30]));  /* !!  ?? */
+  *((ip_t*)(&tmp[8])) = *((ip_t*)(&tmp[26]));
   memcpy(p->data, &tmp[0], p->len);
 }
 
+int xsend(conn_list_t *cl, void *packet)
+{
+  int r;
+  real_packet_t rp;
+
+  do_encrypt((vpn_packet_t*)packet, &rp, cl->key);
+  rp.from = myself->vpn_ip;
+
+  if(debug_lvl > 3)
+    syslog(LOG_ERR, "Sent %d bytes to %lx", rp.len, cl->vpn_ip);
+
+  if((r = send(cl->socket, (char*)&rp, rp.len, 0)) < 0)
+    {
+      syslog(LOG_ERR, "Error sending data: %m");
+      return -1;
+    }
+
+  total_socket_out += r;
+
+  return 0;
+}
+
+/*
+  write as many bytes as possible to the tap
+  device, possibly in multiple turns.
+*/
+int write_n(int fd, void *buf, size_t len)
+{
+  int r, done = 0;
+  
+  do
+    {
+      if((r = write(fd, buf, len)) < 0)
+	return -1;
+      len -= r;
+      buf += r;
+      done += r;
+    } while(len > 0);
+
+  return done;
+}
+
+int xrecv(conn_list_t *cl, void *packet)
+{
+  vpn_packet_t vp;
+  int lenin;
+
+  do_decrypt((real_packet_t*)packet, &vp, cl->key);
+  add_mac_addresses(&vp);
+
+  if((lenin = write_n(tap_fd, &vp, vp.len + 2)) < 0)
+    syslog(LOG_ERR, "Can't write to tap device: %m");
+  else
+    total_tap_out += lenin;
+
+  return 0;
+}
+
+/*
+  add the given packet of size s to the
+  queue q, be it the send or receive queue
+*/
+void add_queue(packet_queue_t **q, void *packet, size_t s)
+{
+  queue_element_t *e;
+
+  if(debug_lvl > 3)
+    syslog(LOG_DEBUG, "packet to queue: %d", s);
+
+  e = xmalloc(sizeof(queue_element_t));
+  e->packet = xmalloc(s);
+  memcpy(e->packet, packet, s);
+  if(!*q)
+    {
+      *q = xmalloc(sizeof(packet_queue_t));
+      (*q)->tail = (*q)->head = NULL;
+    }
+
+  e->prev = (*q)->tail;
+  e->next = NULL;
+  if((*q)->tail)
+    (*q)->tail->next = e;
+
+  (*q)->tail = e;
+  if(!(*q)->head)
+    (*q)->head = e;
+}
+
+/*
+  flush a queue by calling function for
+  each packet, and removing it when that
+  returned a zero exit code
+*/
+void flush_queue(conn_list_t *cl, packet_queue_t *pq,
+		 int (*function)(conn_list_t*,void*))
+{
+  queue_element_t *p, *prev = NULL, *next = NULL;
+
+  for(p = pq->head; p != NULL; )
+    {
+      next = p->next;
+
+cp
+      if(!function(cl, p->packet))
+        {
+          if(prev)
+            prev->next = next;
+          else
+            pq->head = next;
+
+cp
+          free(p->packet);
+cp
+	  free(p);
+cp
+        }
+      else
+        prev = p;
+cp
+
+      p = next;
+    }
+
+  if(debug_lvl > 3)
+    syslog(LOG_DEBUG, "queue flushed");
+}
+
+/*
+  flush the send&recv queues
+  void because nothing goes wrong here, packets
+  remain in the queue if something goes wrong
+*/
+void flush_queues(conn_list_t *cl)
+{
+cp
+  if(cl->sq)
+    {
+      if(debug_lvl > 1)
+	syslog(LOG_DEBUG, "Flushing send queue for " IP_ADDR_S,
+	       IP_ADDR_V(cl->vpn_ip));
+      flush_queue(cl, cl->sq, xsend);
+    }
+cp
+
+  if(cl->rq)
+    {
+      if(debug_lvl > 1)
+	syslog(LOG_DEBUG, "Flushing receive queue for " IP_ADDR_S,
+	       IP_ADDR_V(cl->vpn_ip));
+      flush_queue(cl, cl->rq, xrecv);
+    }
+cp
+}
+
 /*
   send a packet to the given vpn ip.
 */
-int send_packet(ip_t to, real_packet_t *packet)
+int send_packet(ip_t to, vpn_packet_t *packet)
 {
-  int r;
-  conn_list_t const *cl;
+  conn_list_t *cl;
 
   if((cl = lookup_conn(to)) == NULL)
     {
@@ -181,33 +253,46 @@ int send_packet(ip_t to, real_packet_t *packet)
       return -1;
     }
 
+cp
+  if(my_key_expiry <= time(NULL))
+    regenerate_keys();
+
+cp
+  if(!cl->status.dataopen)
+    if(setup_vpn_connection(cl) < 0)
+      return -1;
+cp
+
+  if(!cl->status.validkey)
+    {
+      add_queue(&(cl->sq), packet, packet->len + 2);
+      if(!cl->status.waitingforkey)
+	send_key_request(to);
+      return 0;
+    }
+
+cp
   if(!cl->status.active)
     {
+      add_queue(&(cl->sq), packet, packet->len + 2);
       if(debug_lvl > 1)
-	syslog(LOG_INFO, IP_ADDR_S " is not ready, dropping packet.", IP_ADDR_V(cl->vpn_ip));
+	syslog(LOG_INFO, IP_ADDR_S " is not ready, queueing packet.", IP_ADDR_V(cl->vpn_ip));
       return 0; /* We don't want to mess up, do we? */
     }
 
-  if((r = send(cl->socket, packet, packet->len, 0)) < 0)
-    {
-      syslog(LOG_ERR, "Error sending data: %m");
-      return -1;
-    }
-
-  total_socket_out += r;
-
-  return r;
+  /* can we send it? can we? can we? huh? */
+cp
+  return xsend(cl, packet);
 }
 
-int send_broadcast(conn_list_t *cl, real_packet_t *packet)
+int send_broadcast(conn_list_t *cl, vpn_packet_t *packet)
 {
   conn_list_t *p;
 
   for(p = cl; p != NULL; p = p->next)
     if(send_packet(p->real_ip, packet) < 0)
       {
-	syslog(LOG_ERR, "%d: Could not send a broadcast packet to %08lx (%08lx): %m",
-	       __LINE__,
+	syslog(LOG_ERR, "Could not send a broadcast packet to %08lx (%08lx): %m",
 	       p->vpn_ip, p->real_ip);
 	break; /* FIXME: should retry later, and send a ping over the metaconnection. */
       }
@@ -215,70 +300,9 @@ int send_broadcast(conn_list_t *cl, real_packet_t *packet)
   return 0;
 }
 
-char *hostlookup(unsigned long addr)
-{
-  char *name;
-  struct hostent *host = NULL;
-  struct in_addr in;
-
-  in.s_addr = addr;
-
-  host = gethostbyaddr((char *)&in, sizeof(in), AF_INET);
-
-  if(host)
-    {
-      name = xmalloc(strlen(host->h_name)+20);
-      sprintf(name, "%s (%s)", host->h_name, inet_ntoa(in));
-    }
-  else
-    {
-      name = xmalloc(20);
-      sprintf(name, "%s", inet_ntoa(in));
-    }
-
-  return name;
-}
-
 /*
-  Turn a string into an IP addy with netmask
-  return NULL on failure
+  open the local ethertap device
 */
-ip_mask_t *
-strtoip(const char *str)
-{
-  ip_mask_t *ip;
-  int masker;
-  char *q, *p;
-  struct hostent *h;
-
-  p = str;
-  if((q = strchr(p, '/')))
-    {
-      *q = '\0';
-      q++; /* q now points to netmask part, or NULL if no mask */
-    }
-
-  if(!(h = gethostbyname(p)))
-    {
-      fprintf(stderr, "Error looking up `%s': %s\n", p, sys_errlist[h_errno]);
-      return NULL;
-    }
-
-  masker = 0;
-  if(q)
-    {
-      masker = strtol(q, &p, 10);
-      if(q == p || (*p))
-	return NULL;
-    }
-
-  ip = xmalloc(sizeof(ip_mask_t));
-  ip->ip = ntohl(*((ip_t*)(h->h_addr_list[0])));
-  ip->mask = ~((1 << (32 - masker)) - 1);
-
-  return ip;
-}
-
 int setup_tap_fd(void)
 {
   int nfd;
@@ -300,6 +324,10 @@ int setup_tap_fd(void)
   return 0;
 }
 
+/*
+  set up the socket that we listen on for incoming
+  (tcp) connections
+*/
 int setup_listen_meta_socket(int port)
 {
   int nfd, flags;
@@ -332,7 +360,7 @@ int setup_listen_meta_socket(int port)
 
   if(bind(nfd, (struct sockaddr *)&a, sizeof(struct sockaddr)))
     {
-      syslog(LOG_ERR, "bind: %m");
+      syslog(LOG_ERR, "Can't bind to port %hd/tcp: %m", port);
       return -1;
     }
 
@@ -345,6 +373,10 @@ int setup_listen_meta_socket(int port)
   return nfd;
 }
 
+/*
+  setup the socket for incoming encrypted
+  data (the udp part)
+*/
 int setup_vpn_in_socket(int port)
 {
   int nfd, flags;
@@ -377,13 +409,16 @@ int setup_vpn_in_socket(int port)
 
   if(bind(nfd, (struct sockaddr *)&a, sizeof(struct sockaddr)))
     {
-      syslog(LOG_ERR, "bind: %m");
+      syslog(LOG_ERR, "Can't bind to port %hd/udp: %m", port);
       return -1;
     }
 
   return nfd;
 }
 
+/*
+  setup an outgoing meta (tcp) socket
+*/
 int setup_outgoing_meta_socket(conn_list_t *cl)
 {
   int flags;
@@ -391,7 +426,7 @@ int setup_outgoing_meta_socket(conn_list_t *cl)
   config_t const *cfg;
 
   if((cfg = get_config_val(upstreamport)) == NULL)
-    cl->port = 24;
+    cl->port = 655;
   else
     cl->port = cfg->data.val;
 
@@ -426,6 +461,13 @@ int setup_outgoing_meta_socket(conn_list_t *cl)
   return 0;
 }
 
+/*
+  setup an outgoing connection. It's not
+  necessary to also open an udp socket as
+  well, because the other host will initiate
+  an authentication sequence during which
+  we will do just that.
+*/
 int setup_outgoing_connection(ip_t ip)
 {
   conn_list_t *ncn;
@@ -448,6 +490,9 @@ int setup_outgoing_connection(ip_t ip)
   return 0;
 }
 
+/*
+  set up the local sockets (listen only)
+*/
 int setup_myself(void)
 {
   config_t const *cfg;
@@ -464,7 +509,7 @@ int setup_myself(void)
   myself->vpn_mask = cfg->data.ip->mask;
 
   if(!(cfg = get_config_val(listenport)))
-    myself->port = 24;
+    myself->port = 655;
   else
     myself->port = cfg->data.val;
 
@@ -488,6 +533,9 @@ int setup_myself(void)
   return 0;
 }
 
+/*
+  setup all initial network connections
+*/
 int setup_network_connections(void)
 {
   config_t const *cfg;
@@ -517,31 +565,34 @@ RETSIGTYPE
 sigalrm_handler(int a)
 {
   config_t const *cfg;
+  static int seconds_till_retry;
 
-  if((cfg = get_config_val(upstreamip)) == NULL)
-    {
-      syslog(LOG_ERR, "Serious programming error. line %d of %s.", __LINE__, __FILE__);
-      return;
-    }
+  cfg = get_config_val(upstreamip);
 
   if(!setup_outgoing_connection(cfg->data.ip->ip))
-    signal(SIGALRM, SIG_IGN);
+    {
+      signal(SIGALRM, SIG_IGN);
+      seconds_till_retry = 5;
+    }
   else
     {
       signal(SIGALRM, sigalrm_handler);
-      alarm(5);
-      syslog(LOG_ERR, "Still failed to connect to other. Will retry in 5 seconds.");
+      seconds_till_retry += 5;
+      alarm(seconds_till_retry);
+      syslog(LOG_ERR, "Still failed to connect to other. Will retry in %d seconds.",
+	     seconds_till_retry);
     }
 }
 
+/*
+  close all open network connections
+*/
 void close_network_connections(void)
 {
   conn_list_t *p;
 
   for(p = conn_list; p != NULL; p = p->next)
     {
-      send_termreq(p);
-      
       if(p->status.active)
 	{
 	  shutdown(p->socket, 0); /* No more receptions */
@@ -549,7 +600,8 @@ void close_network_connections(void)
 	}
       if(p->status.meta)
 	{
-	  shutdown(p->socket, 0); /* No more receptions */
+	  send_termreq(p);
+	  shutdown(p->meta_socket, 0); /* No more receptions */
           close(p->meta_socket);
         }
     }
@@ -568,15 +620,21 @@ void close_network_connections(void)
   return;
 }
 
+/*
+  create a data (udp) socket
+*/
 int setup_vpn_connection(conn_list_t *cl)
 {
   int nfd, flags;
   struct sockaddr_in a;
 
+  if(debug_lvl > 1)
+    syslog(LOG_DEBUG, "Opening UDP socket to " IP_ADDR_S, IP_ADDR_V(cl->real_ip));
+
   nfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
   if(nfd == -1)
     {
-      syslog(LOG_ERR, "%d: Creating socket failed: %m", __LINE__);
+      syslog(LOG_ERR, "Creating data socket failed: %m");
       return -1;
     }
 
@@ -586,7 +644,7 @@ int setup_vpn_connection(conn_list_t *cl)
 
   if(connect(nfd, (struct sockaddr *)&a, sizeof(a)) == -1)
     {
-      syslog(LOG_ERR, "%d: %08lx:%d: %m", __LINE__, ntohs(cl->real_ip),
+      syslog(LOG_ERR, "Create connection to %08lx:%d failed: %m", ntohs(cl->real_ip),
 	     cl->port);
       return -1;
     }
@@ -594,15 +652,20 @@ int setup_vpn_connection(conn_list_t *cl)
   flags = fcntl(nfd, F_GETFL);
   if(fcntl(nfd, F_SETFL, flags | O_NONBLOCK) < 0)
     {
-      syslog(LOG_ERR, "%d: fcntl: %m", __LINE__);
+      syslog(LOG_ERR, "This is a bug: %s:%d: %d:%m", __FILE__, __LINE__, nfd);
       return -1;
     }
 
   cl->socket = nfd;
+  cl->status.dataopen = 1;
 
   return 0;
 }
 
+/*
+  handle an incoming tcp connect call and open
+  a connection to it.
+*/
 conn_list_t *create_new_connection(int sfd)
 {
   conn_list_t *p;
@@ -633,6 +696,9 @@ conn_list_t *create_new_connection(int sfd)
   return p;
 }
 
+/*
+  put all file descriptors in an fd_set array
+*/
 void build_fdset(fd_set *fs)
 {
   conn_list_t *p;
@@ -652,45 +718,70 @@ void build_fdset(fd_set *fs)
   FD_SET(tap_fd, fs);
 }
 
+/*
+  receive incoming data from the listening
+  udp socket and write it to the ethertap
+  device after being decrypted
+*/
 int handle_incoming_vpn_data(conn_list_t *cl)
 {
   real_packet_t rp;
-  vpn_packet_t vp;
   int lenin;
   int x, l = sizeof(x);
+  conn_list_t *f;
 
   if(getsockopt(cl->socket, SOL_SOCKET, SO_ERROR, &x, &l) < 0)
     {
-      syslog(LOG_ERR, "%d: getsockopt failed: %m", __LINE__);
+      syslog(LOG_ERR, "This is a bug: %s:%d: %d:%m", __FILE__, __LINE__, cl->socket);
       return -1;
     }
   if(x)
     {
-      syslog(LOG_ERR, "%d: Socket error: %s", __LINE__, sys_errlist[x]);
+      syslog(LOG_ERR, "Incoming data socket error: %s", sys_errlist[x]);
       return -1;
     }
 
   rp.len = -1;
   lenin = recvfrom(cl->socket, &rp, MTU, 0, NULL, NULL);
-  if(lenin <= 0) /* Waarschijnlijk is de andere kant er mee gekapt */
+  if(lenin <= 0)
     {
-      syslog(LOG_ERR, "%d: Receive: %m.\n", __LINE__);
+      syslog(LOG_ERR, "Receiving data failed: %m");
       return -1;
     }
   total_socket_in += lenin;
   if(rp.len >= 0)
     {
-      do_decrypt(&rp, &vp);
-      add_mac_addresses(&vp);
-      if((lenin = write(tap_fd, &vp, vp.len + 2)) < 0)
-	syslog(LOG_ERR, "%d: write: %m", __LINE__);
+      f = lookup_conn(rp.from);
+      if(debug_lvl > 2)
+	syslog(LOG_DEBUG, "packet from " IP_ADDR_S " (len %d)",
+	       IP_ADDR_V(rp.from), rp.len);
+      if(!f)
+	{
+	  syslog(LOG_ERR, "Got packet from unknown source " IP_ADDR_S,
+		 IP_ADDR_V(rp.from));
+	  return -1;
+	}
+
+      if(f->status.validkey)
+	xrecv(f, &rp);
       else
-	total_tap_out += lenin;
+	{
+	  add_queue(&(f->rq), &rp, rp.len);
+	  if(!cl->status.waitingforkey)
+	    send_key_request(rp.from);
+	}
+
+      if(my_key_expiry <= time(NULL))
+	regenerate_keys();
     }
 
   return 0;
 }
 
+/*
+  terminate a connection and notify the other
+  end before closing the sockets
+*/
 void terminate_connection(conn_list_t *cl)
 {
   if(cl->status.remove)
@@ -718,6 +809,10 @@ void terminate_connection(conn_list_t *cl)
   cl->status.remove = 1;
 }
 
+/*
+  send out a ping request to all active
+  connections
+*/
 int send_broadcast_ping(void)
 {
   conn_list_t *p;
@@ -738,9 +833,15 @@ int send_broadcast_ping(void)
 	}
     }
 
+  last_ping_time = time(NULL);
+
   return 0;
 }
 
+/*
+  end all connections that did not respond
+  to the ping probe in time
+*/
 int check_dead_connections(void)
 {
   conn_list_t *p;
@@ -761,6 +862,10 @@ int check_dead_connections(void)
   return 0;
 }
 
+/*
+  accept a new tcp connect and create a
+  new connection
+*/
 int handle_new_meta_connection(conn_list_t *cl)
 {
   conn_list_t *ncn;
@@ -769,7 +874,7 @@ int handle_new_meta_connection(conn_list_t *cl)
 
   if((nfd = accept(cl->meta_socket, &client, &len)) < 0)
     {
-      syslog(LOG_ERR, "accept: %m");
+      syslog(LOG_ERR, "Accepting a new connection failed: %m");
       return -1;
     }
 
@@ -788,6 +893,9 @@ int handle_new_meta_connection(conn_list_t *cl)
   return 0;
 }
 
+/*
+  dispatch any incoming meta requests
+*/
 int handle_incoming_meta_data(conn_list_t *cl)
 {
   int x, l = sizeof(x), lenin;
@@ -796,18 +904,18 @@ int handle_incoming_meta_data(conn_list_t *cl)
   
   if(getsockopt(cl->meta_socket, SOL_SOCKET, SO_ERROR, &x, &l) < 0)
     {
-      syslog(LOG_ERR, "%d: getsockopt failed: %m", __LINE__);
+      syslog(LOG_ERR, "This is a bug: %s:%d: %d:%m", __FILE__, __LINE__, cl->meta_socket);
       return -1;
     }
   if(x)
     {
-      syslog(LOG_ERR, "Socket error: %s", __LINE__, sys_errlist[x]);
+      syslog(LOG_ERR, "Metadata socket error: %s", sys_errlist[x]);
       return -1;
     }
 
   if((lenin = recv(cl->meta_socket, &tmp, sizeof(tmp), 0)) <= 0)
     {
-      syslog(LOG_ERR, "%d: receive: %m", __LINE__);
+      syslog(LOG_ERR, "Receive failed: %m");
       return -1;
     }
 
@@ -817,10 +925,7 @@ int handle_incoming_meta_data(conn_list_t *cl)
     syslog(LOG_DEBUG, "got request %d", request);
 
   if(request_handlers[request] == NULL)
-    {
-      if(debug_lvl > 0)
-	syslog(LOG_ERR, "Unknown request %d.", request);
-    }
+    syslog(LOG_ERR, "Unknown request %d.", request);
   else
     if(request_handlers[request](cl, tmp, lenin) < 0)
       return -1;
@@ -828,6 +933,10 @@ int handle_incoming_meta_data(conn_list_t *cl)
   return 0;
 }
 
+/*
+  check all connections to see if anything
+  happened on their sockets
+*/
 void check_network_activity(fd_set *f)
 {
   conn_list_t *p;
@@ -837,6 +946,7 @@ void check_network_activity(fd_set *f)
     {
       if(p->status.remove)
 	continue;
+cp
       if(p->status.active)
 	if(FD_ISSET(p->socket, f))
 	  {
@@ -847,10 +957,11 @@ void check_network_activity(fd_set *f)
 	      I've once got here when it said `No route to host'.
 	    */
 	    getsockopt(p->socket, SOL_SOCKET, SO_ERROR, &x, &l);
-	    syslog(LOG_ERR, "%d: Data socket error: %s", __LINE__, sys_errlist[x]);
+	    syslog(LOG_ERR, "Outgoing data socket error: %s", sys_errlist[x]);
 	    terminate_connection(p);
 	    return;
 	  }  
+cp
       if(p->status.meta)
 	if(FD_ISSET(p->meta_socket, f))
 	  if(handle_incoming_meta_data(p) < 0)
@@ -858,17 +969,24 @@ void check_network_activity(fd_set *f)
 	      terminate_connection(p);
 	      return;
 	    } 
+cp
     }
   
+cp
   if(FD_ISSET(myself->socket, f))
     handle_incoming_vpn_data(myself);
+cp
   if(FD_ISSET(myself->meta_socket, f))
     handle_new_meta_connection(myself);
+cp
 }
 
+/*
+  read, encrypt and send data that is
+  available through the ethertap device
+*/
 void handle_tap_input(void)
 {
-  real_packet_t rp;
   vpn_packet_t vp;
   ip_t from, to;
   int ether_type, lenin;
@@ -879,66 +997,89 @@ void handle_tap_input(void)
       syslog(LOG_ERR, "Error while reading from tapdevice: %m");
       return;
     }
-  
+
   total_tap_in += lenin;
-  
+
   ether_type = ntohs(*((unsigned short*)(&vp.data[12])));
   if(ether_type != 0x0800)
     {
-      if(debug_lvl >= 1)
+      if(debug_lvl > 0)
 	syslog(LOG_INFO, "Non-IP ethernet frame %04x from " MAC_ADDR_S,
 	       ether_type, MAC_ADDR_V(vp.data[6]));
       return;
     }
   
+  if(lenin < 32)
+    {
+      if(debug_lvl > 0)
+	syslog(LOG_INFO, "Dropping short packet");
+      return;
+    }
+
   from = ntohl(*((unsigned long*)(&vp.data[26])));
   to = ntohl(*((unsigned long*)(&vp.data[30])));
-  
-  if(debug_lvl > 1)
+
+  if(debug_lvl > 2)
     syslog(LOG_DEBUG, "An IP packet (%04x) for " IP_ADDR_S " from " IP_ADDR_S,
 	   ether_type, IP_ADDR_V(to), IP_ADDR_V(from));
-  if(debug_lvl > 2)
+  if(debug_lvl > 3)
     syslog(LOG_DEBUG, MAC_ADDR_S " to " MAC_ADDR_S,
 	   MAC_ADDR_V(vp.data[0]), MAC_ADDR_V(vp.data[6]));
   
   vp.len = (length_t)lenin - 2;
+cp
   strip_mac_addresses(&vp);
-  do_encrypt(&vp, &rp);
-  send_packet(to, &rp);
+cp
+  send_packet(to, &vp);
+cp
 }
 
+/*
+  this is where it al happens...
+*/
 void main_loop(void)
 {
   fd_set fset;
   struct timeval tv;
   int r;
 
+  last_ping_time = time(NULL);
+
   for(;;)
     {
       tv.tv_sec = timeout;
       tv.tv_usec = 0;
 
+cp
       prune_conn_list();
+cp
       build_fdset(&fset);
+cp
 
       if((r = select(FD_SETSIZE, &fset, NULL, NULL, &tv)) < 0)
         {
 	  if(errno == EINTR) /* because of alarm */
 	    continue;
-          syslog(LOG_ERR, "Error while waiting for input: %m\n");
+          syslog(LOG_ERR, "Error while waiting for input: %m");
           return;
         }
-      if(r == 0)     /* Timeout... hm... something might be wrong. */
+cp
+
+      if(r == 0 || last_ping_time + timeout < time(NULL))
+	/* Timeout... hm... something might be wrong. */
 	{
 	  check_dead_connections();
 	  send_broadcast_ping();
 	  continue;
 	}
 
+cp
       check_network_activity(&fset);
 
+cp
       /* local tap data */
       if(FD_ISSET(tap_fd, &fset))
 	handle_tap_input();
+cp
     }
 }
