@@ -20,6 +20,8 @@
 #ifndef __TINC_NET_H__
 #define __TINC_NET_H__
 
+#include <sys/time.h>
+
 #include "config.h"
 #include "conf.h"
 
@@ -48,13 +50,14 @@ typedef unsigned long ip_t;
 typedef short length_t;
 
 typedef struct vpn_packet_t {
-  length_t len;
+  length_t len;		/* the actual number of bytes in the `data' field */
   unsigned char data[MAXSIZE];
 } vpn_packet_t;
 
 typedef struct real_packet_t {
-  length_t len;
-  vpn_packet_t data;
+  length_t len;		/* the length of the entire packet */
+  ip_t from;		/* where the packet came from */
+  vpn_packet_t data;	/* encrypted vpn_packet_t */
 } real_packet_t;
 
 typedef struct passphrase_t {
@@ -72,9 +75,29 @@ typedef struct status_bits_t {
   int termreq:1;                   /* the termination of this connection was requested */
   int remove:1;                    /* Set to 1 if you want this connection removed */
   int timeout:1;                   /* 1 if gotten timeout */
-  int unused:25;
+  int validkey:1;                  /* 1 if we currently have a valid key for him */
+  int waitingforkey:1;             /* 1 if we already sent out a request */
+  int dataopen:1;                  /* 1 if we have a valid UDP connection open */
+  int unused:22;
 } status_bits_t;
 
+typedef struct queue_element_t {
+  void *packet;
+  struct queue_element_t *prev;
+  struct queue_element_t *next;
+} queue_element_t;
+
+typedef struct packet_queue_t {
+  queue_element_t *head;
+  queue_element_t *tail;
+} packet_queue_t;
+
+typedef struct enc_key_t {
+  int length;
+  char *key;
+  time_t expiry;
+} enc_key_t;
+
 typedef struct conn_list_t {
   ip_t vpn_ip;                     /* his vpn ip */
   ip_t vpn_mask;                   /* his vpn network address */
@@ -86,7 +109,12 @@ typedef struct conn_list_t {
   unsigned char protocol_version;  /* used protocol */
   status_bits_t status;            /* status info */
   passphrase_t *pp;                /* encoded passphrase */
-  int nr;                          /* serial number when calculating key */
+  packet_queue_t *sq;              /* pending outgoing packets */
+  packet_queue_t *rq;              /* pending incoming packets (they have no
+				      valid key to be decrypted with) */
+  enc_key_t *public_key;           /* the other party's public key */
+  enc_key_t *key;                  /* encrypt with this key */
+  struct conn_list_t *nexthop;     /* nearest meta-hop in this direction */
   struct conn_list_t *next;        /* after all, it's a list of connections */
 } conn_list_t;
 
@@ -100,13 +128,13 @@ extern int total_socket_out;
 extern conn_list_t *conn_list;
 extern conn_list_t *myself;
 
-extern int send_packet(ip_t, real_packet_t *);
-extern int send_broadcast(conn_list_t *, real_packet_t *);
+extern int send_packet(ip_t, vpn_packet_t *);
+extern int send_broadcast(conn_list_t *, vpn_packet_t *);
 extern int setup_network_connections(void);
 extern void close_network_connections(void);
 extern void main_loop(void);
 extern int setup_vpn_connection(conn_list_t *);
 extern void terminate_connection(conn_list_t *);
-extern ip_mask_t *strtoip(const char*);
+extern void flush_queues(conn_list_t*);
 
 #endif /* __TINC_NET_H__ */
