@@ -25,7 +25,8 @@
 enum {
   PROT_RESERVED = 0,                 /* reserved: do not use. */
   PROT_NOT_IN_USE,
-  PROT_TOO_OLD,
+  PROT_TOO_OLD = 2,
+  PROT_3,
   PROT_CURRENT,                      /* protocol currently in use */
 };
 
@@ -41,6 +42,7 @@ enum {
   AUTH_C_RACK,          /* waiting for ack */
   TERMREQ = 30,         /* terminate connection */
   PINGTIMEOUT,          /* terminate due to ping t.o. */
+  DEL_HOST,		/* forward a termreq to others */
   PING = 40,            /* ping */
   PONG,
   ADD_HOST = 60,        /* Add new given host to connection list */
@@ -52,6 +54,9 @@ enum {
   CALCULATE = 100,      /* calculate the following numer^privkey and send me the result */  
   CALC_RES,             /* result of the above */
   ALMOST_KEY,           /* this number^privkey is the shared key */
+  REQ_KEY = 160,        /* request public key */
+  ANS_KEY,              /* answer to such request */
+  KEY_CHANGED,		/* public key has changed */
 };
 
 typedef struct add_host_t {
@@ -87,12 +92,33 @@ typedef struct public_key_t {
   char key;
 } public_key_t;
 
+typedef struct key_req_t {
+  unsigned char type;
+  ip_t from;
+  ip_t to;
+  time_t expiry;
+  short int len; /* 0 if requesting */
+  char key;
+} key_req_t;
+
+typedef struct key_changed_t {
+  unsigned char type;
+  ip_t from;
+} key_changed_t;
+
+typedef struct del_host_t {
+  unsigned char type;
+  ip_t vpn_ip;
+} del_host_t;
+
 extern int (*request_handlers[256])(conn_list_t*, unsigned char*, int);
 
 extern int send_ping(conn_list_t*);
 extern int send_basic_info(conn_list_t *);
 extern int send_termreq(conn_list_t *);
 extern int send_timeout(conn_list_t *);
+extern int send_key_request(ip_t);
+extern void send_key_changed2(void);
 
 #endif /* __TINC_PROTOCOL_H__ */
 
