@@ -33,11 +33,13 @@
 #endif
 
 #include <pidfile.h>
+#include <utils.h>
 #include <xalloc.h>
 
 #include "conf.h"
 #include "encr.h"
 #include "net.h"
+#include "netutl.h"
 
 /* The name this program was run with. */
 char *program_name;
@@ -180,7 +182,12 @@ int detach(void)
   chdir("/"); /* avoid keeping a mointpoint busy */
 
   openlog(identname, LOG_CONS | LOG_PID, LOG_DAEMON);
-  syslog(LOG_NOTICE, "tincd %s starting", VERSION);
+
+  if(debug_lvl > 1)
+    syslog(LOG_NOTICE, "tincd %s (%s %s) starting, debug level %d.",
+	   VERSION, __DATE__, __TIME__, debug_lvl);
+  else
+    syslog(LOG_NOTICE, "tincd %s starting, debug level %d.", VERSION, debug_lvl);
 
   return 0;
 }
@@ -191,9 +198,11 @@ int detach(void)
 void cleanup_and_exit(int c)
 {
   close_network_connections();
+
   if(debug_lvl > 0)
     syslog(LOG_INFO, "Total bytes written: tap %d, socket %d; bytes read: tap %d, socket %d.",
 	   total_tap_out, total_socket_out, total_tap_in, total_socket_in);
+
   closelog();
   kill(ppid, SIGTERM);
   exit(c);
@@ -257,26 +266,30 @@ void make_names(void)
     {
       if(netname)
 	{
-	  configfilename = xmalloc(strlen(netname)+22);
-	  sprintf(configfilename, "/etc/tinc/%s/tincd.conf", netname);
+	  configfilename = xmalloc(strlen(netname)+18+strlen(CONFDIR));
+	  sprintf(configfilename, "%s/tinc/%s/tincd.conf", CONFDIR, netname);
 	}
       else
-	configfilename = "/etc/tinc/tincd.conf";
+	{
+	  configfilename = xmalloc(17+strlen(CONFDIR));
+	  sprintf(configfilename, "%s/tinc/tincd.conf", CONFDIR);
+	}
     }
   
   if(netname)
     {
       pidfilename = xmalloc(strlen(netname)+20);
       sprintf(pidfilename, "/var/run/tincd.%s.pid", netname);
-      confbase = xmalloc(strlen(netname)+12);
-      sprintf(confbase, "/etc/tinc/%s/", netname);
+      confbase = xmalloc(strlen(netname)+8+strlen(CONFDIR));
+      sprintf(confbase, "%s/tinc/%s/", CONFDIR, netname);
       identname = xmalloc(strlen(netname)+7);
       sprintf(identname, "tincd.%s", netname);
     }
   else
     {
       pidfilename = "/var/run/tincd.pid";
-      confbase = "/etc/tinc/";
+      confbase = xmalloc(7+strlen(CONFDIR));
+      sprintf(confbase, "%s/tinc/", CONFDIR);
       identname = "tincd";
     }
 }
@@ -352,14 +365,27 @@ sigquit_handler(int a)
   cleanup_and_exit(0);
 }
 
+RETSIGTYPE
+sigsegv_square(int a)
+{
+  syslog(LOG_NOTICE, "Got another SEGV signal: not restarting");
+  exit(0);
+}
+
 RETSIGTYPE
 sigsegv_handler(int a)
 {
-  syslog(LOG_NOTICE, "Got SEGV signal: re-executing self");
+  if(cp_file)
+    syslog(LOG_NOTICE, "Got SEGV signal after %s line %d. Trying to re-execute.",
+	   cp_file, cp_line);
+  else
+    syslog(LOG_NOTICE, "Got SEGV signal; trying to re-execute.");
+
+  signal(SIGSEGV, sigsegv_square);
+
   close_network_connections();
   remove_pid(pidfilename);
   execvp(g_argv[0], g_argv);
-  /*  cleanup_and_exit(0); */
 }
 
 RETSIGTYPE
@@ -380,6 +406,20 @@ sigint_handler(int a)
   cleanup_and_exit(0);
 }
 
+RETSIGTYPE
+sigusr1_handler(int a)
+{
+  dump_conn_list();
+}
+
+RETSIGTYPE
+sigusr2_handler(int a)
+{
+  if(debug_lvl > 1)
+    syslog(LOG_NOTICE, "Forcing new keys");
+  regenerate_keys();
+}
+
 void
 setup_signals(void)
 {
@@ -394,6 +434,8 @@ setup_signals(void)
   signal(SIGPIPE, SIG_IGN);
   if(signal(SIGINT, SIG_IGN) != SIG_ERR)
     signal(SIGINT, sigint_handler);
+  signal(SIGUSR1, sigusr1_handler);
+  signal(SIGUSR2, sigusr2_handler);
 }
 
 RETSIGTYPE parent_exit(int a)
